import dataclasses as dc
import multiprocessing as mp
import multiprocessing.popen_spawn_posix
import sys
import typing as ta

from .. import check
from .. import libc


T = ta.TypeVar('T')


##


@dc.dataclass(frozen=True, kw_only=True)
class SpawnExtras:
    pass_fds: ta.AbstractSet[int] | None = None
    deathsig: int | None = None


class ExtrasSpawnPosixPopen(mp.popen_spawn_posix.Popen):
    def __init__(self, process_obj: 'ExtrasSpawnProcess', *, extras: SpawnExtras) -> None:
        self.__extras = extras
        self.__pass_fds = extras.pass_fds

        super().__init__(process_obj)

    def _launch(self, process_obj: 'ExtrasSpawnProcess') -> None:
        if self.__pass_fds:
            for fd in self.__pass_fds:
                self.duplicate_for_child(fd)
            self._extra_fds = None

        super()._launch(process_obj)  # type: ignore  # noqa


class ExtrasSpawnProcess(mp.context.SpawnProcess):
    def __init__(self, *args: ta.Any, extras: SpawnExtras, **kwargs: ta.Any) -> None:
        self.__extras = extras

        super().__init__(*args, **kwargs)

    def _Popen(self, process_obj: 'ExtrasSpawnProcess') -> ExtrasSpawnPosixPopen:  # type: ignore  # noqa
        return ExtrasSpawnPosixPopen(
            check.isinstance(process_obj, ExtrasSpawnProcess),
            extras=self.__extras,
        )

    def run(self) -> None:
        if self.__extras.deathsig is not None and sys.platform == 'linux':
            libc.prctl(libc.PR_SET_PDEATHSIG, self.__extras.deathsig, 0, 0, 0, 0)

        super().run()


class ExtrasSpawnContext(mp.context.SpawnContext):
    def __init__(self, extras: SpawnExtras = SpawnExtras()) -> None:
        self.__extras = extras

        super().__init__()

    def Process(self, *args: ta.Any, **kwargs: ta.Any):  # type: ignore  # noqa
        return ExtrasSpawnProcess(*args, extras=self.__extras, **kwargs)



import functools
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang


##


class LifecycleStateError(Exception):
    pass


@ta.final
@dc.dataclass(frozen=True, eq=False)
@functools.total_ordering
class LifecycleState(lang.Final):
    name: str
    phase: int
    is_failed: bool

    def __lt__(self, other):
        return self.phase < check.isinstance(other, LifecycleState).phase

    def __le__(self, other):
        return self.phase <= check.isinstance(other, LifecycleState).phase


class LifecycleStates(lang.Namespace):
    NEW = LifecycleState('NEW', 0, False)

    CONSTRUCTING = LifecycleState('CONSTRUCTING', 1, False)
    FAILED_CONSTRUCTING = LifecycleState('FAILED_CONSTRUCTING', 2, True)
    CONSTRUCTED = LifecycleState('CONSTRUCTED', 3, False)

    STARTING = LifecycleState('STARTING', 4, False)
    FAILED_STARTING = LifecycleState('FAILED_STARTING', 5, True)
    STARTED = LifecycleState('STARTED', 6, False)

    STOPPING = LifecycleState('STOPPING', 7, False)
    FAILED_STOPPING = LifecycleState('FAILED_STOPPING', 8, True)
    STOPPED = LifecycleState('STOPPED', 9, False)

    DESTROYING = LifecycleState('DESTROYING', 10, False)
    FAILED_DESTROYING = LifecycleState('FAILED_DESTROYING', 11, True)
    DESTROYED = LifecycleState('DESTROYED', 12, False)



from .... import check
from ..._internals import STD_PARAMS_ATTR
from ..processing.base import Processor
from ..processing.phases import ProcessorPhase
from ..processing.registry import register_processor_type


##


@register_processor_type(phase=ProcessorPhase.BOOTSTRAP)
class ParamsProcessor(Processor):
    def check(self) -> None:
        check.in_(STD_PARAMS_ATTR, self._ctx.cls.__dict__)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ....io.pipelines.bytes.buffers import OutboundBytesBufferIoPipelineHandler
from ....io.pipelines.core import IoPipeline
from ....io.pipelines.core import IoPipelineHandler
from ....io.pipelines.flow.stub import StubIoPipelineFlowService
from ....io.pipelines.handlers.logs import LoggingIoPipelineHandler
from ....io.pipelines.ssl.handlers import SslIoPipelineHandler
from ....lite.abstract import Abstract
from ...clients.base import BaseHttpClient
from ...clients.base import HttpClientRequest
from ...headers import HttpHeaders
from ...pipelines.clients.clients import IoPipelineHttpClientHandler
from ...pipelines.clients.requests import IoPipelineHttpRequestCompressor
from ...pipelines.clients.requests import IoPipelineHttpRequestEncoder
from ...pipelines.clients.responses import IoPipelineHttpResponseAggregatorDecoder
from ...pipelines.clients.responses import IoPipelineHttpResponseDecoder
from ...pipelines.clients.responses import IoPipelineHttpResponseDecompressor
from ...pipelines.requests import FullIoPipelineHttpRequest


##


class BaseIoPipelineHttpClient(BaseHttpClient, Abstract):
    def __init__(self, **pipeline_kwargs: ta.Any) -> None:
        super().__init__()

        self._pipeline_kwargs = pipeline_kwargs

    #

    @dc.dataclass(frozen=True)
    class ParsedUrl:
        host: str
        port: int
        path: str

        is_ssl: bool = False

    @classmethod
    def parse_url(cls, url: str) -> ParsedUrl:
        # Parse URL (very simple - just extract host and path)
        if url.startswith('http://'):
            is_ssl = False
            url_without_scheme = url[7:]
        elif url.startswith('https://'):
            is_ssl = True
            url_without_scheme = url[8:]
        else:
            raise ValueError(url)

        if '/' in url_without_scheme:
            host, path = url_without_scheme.split('/', 1)
            path = '/' + path
        else:
            host = url_without_scheme
            path = '/'

        # Extract port if present
        if ':' in host:
            host, port_str = host.split(':', 1)
            port = int(port_str)
        else:
            if is_ssl:
                port = 443
            else:
                port = 80

        return cls.ParsedUrl(
            host,
            port,
            path,
            is_ssl=is_ssl,
        )

    #

    def _build_pipeline_spec(
            self,

            outermost_handlers: ta.Optional[ta.Sequence[IoPipelineHandler]] = None,
            innermost_handlers: ta.Optional[ta.Sequence[IoPipelineHandler]] = None,

            with_logging: bool = False,

            with_ssl: bool = False,
            ssl_kwargs: ta.Optional[ta.Mapping[str, ta.Any]] = None,

            with_flow: bool = False,
            flow_auto_read: bool = False,

            raise_immediately: bool = False,
    ) -> IoPipeline.Spec:
        return IoPipeline.Spec(
            [
                *(outermost_handlers or []),

                *([LoggingIoPipelineHandler()] if with_logging else []),

                *([OutboundBytesBufferIoPipelineHandler()] if with_flow else []),

                *([SslIoPipelineHandler(**(ssl_kwargs or {}))] if with_ssl else []),

                IoPipelineHttpResponseDecoder(),
                IoPipelineHttpResponseDecompressor(),
                IoPipelineHttpResponseAggregatorDecoder(),

                IoPipelineHttpRequestEncoder(),
                IoPipelineHttpRequestCompressor(),

                IoPipelineHttpClientHandler(),

                *(innermost_handlers or []),
            ],

            config=IoPipeline.Config.DEFAULT.update(
                raise_immediately=raise_immediately,
            ),

            services=[
                *([StubIoPipelineFlowService(auto_read=flow_auto_read)] if with_flow else []),
            ],
        )

    #

    @dc.dataclass(frozen=True)
    class _PreparedRequest:
        parsed_url: 'BaseIoPipelineHttpClient.ParsedUrl'
        full_request: FullIoPipelineHttpRequest
        pipeline_spec: IoPipeline.Spec

    def _prepare_request(
            self,
            req: HttpClientRequest,
            **pipeline_kwargs: ta.Any,
    ) -> _PreparedRequest:
        parsed_url = self.parse_url(req.url)

        full_request = FullIoPipelineHttpRequest.simple(
            parsed_url.host,
            parsed_url.path,
            method=req.method_or_default,
            headers=HttpHeaders.of(req.headers_).update(
                ('User-Agent', 'omlish-http-client/0.1'),
                if_present='skip',
            ),
        )

        pipeline_spec = self._build_pipeline_spec(
            **(dict(  # type: ignore[arg-type]
                with_ssl=True,
                ssl_kwargs=dict(
                    server_side=False,
                    server_hostname=parsed_url.host,
                ),
            ) if parsed_url.is_ssl else {}),

            **{
                **self._pipeline_kwargs,
                **pipeline_kwargs,
            },
        )

        return self._PreparedRequest(
            parsed_url,
            full_request,
            pipeline_spec,
        )



# ruff: noqa: UP006 UP043 UP045
# @omlish-lite
import threading
import typing as ta


T = ta.TypeVar('T')


##


class SyncBufferRelay(ta.Generic[T]):
    def __init__(
            self,
            *,
            wake_fn: ta.Callable[[], None],
    ) -> None:
        super().__init__()

        self._wake_fn = wake_fn

        self._lock = threading.Lock()
        self._buffer: SyncBufferRelay._Buffer = SyncBufferRelay._Buffer()

    class _Buffer:
        def __init__(self) -> None:
            self.lst: list = []
            self.age = 0

        def __repr__(self) -> str:
            return f'{self.__class__.__qualname__}({self.lst!r}, age={self.age!r})'

    def push(self, *vs: T) -> None:
        with self._lock:
            buf = self._buffer
            needs_wake = not buf.age
            buf.lst.extend(vs)
            buf.age += 1
        if needs_wake:
            self._wake_fn()

    def swap(self) -> ta.Sequence[T]:
        with self._lock:
            buf, self._buffer = self._buffer, SyncBufferRelay._Buffer()
        return buf.lst



# ruff: noqa: UP006 UP037 UP045
# @omlish-lite
import dataclasses as dc
import typing as ta

from ..core import IoPipelineHandler
from ..core import IoPipelineHandlerContext


##


class FeedbackInboundIoPipelineHandler(IoPipelineHandler):
    @ta.final
    @dc.dataclass(frozen=True)
    class FeedbackMessages:
        msgs: ta.Iterable[ta.Any]

        _handler: ta.Optional['FeedbackInboundIoPipelineHandler'] = None

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if not isinstance(msg, FeedbackInboundIoPipelineHandler.FeedbackMessages):
            ctx.feed_in(msg)
            return

        if not (msg._handler is None or msg._handler is self):  # noqa
            ctx.feed_in(msg)
            return

        for m in msg.msgs:
            ctx.feed_out(m)

    def wrap(self, *msgs: ta.Any) -> FeedbackMessages:
        return FeedbackInboundIoPipelineHandler.FeedbackMessages(msgs, self)



# ruff: noqa: UP045
# @omlish-lite
import abc
import typing as ta

from ....lite.abstract import Abstract
from ..core import IoPipelineHandler


##


class InboundBytesBufferingIoPipelineHandler(IoPipelineHandler, Abstract):
    @abc.abstractmethod
    def inbound_buffered_bytes(self) -> ta.Optional[int]:
        """Returning `None` denotes currently unknown/unanswerable."""

        raise NotImplementedError


class OutboundBytesBufferingIoPipelineHandler(IoPipelineHandler, Abstract):
    @abc.abstractmethod
    def outbound_buffered_bytes(self) -> ta.Optional[int]:
        """Returning `None` denotes currently unknown/unanswerable."""

        raise NotImplementedError



import typing as ta

from ... import dataclasses as dc
from ._compat import compat
from .types import ImmutableMultiDict


##


@compat
@dc.dataclass(frozen=True, kw_only=True)
class Request:
    path: str
    method: str

    form: ImmutableMultiDict[str, str] | None = None

    def get_json(self) -> ta.Any:
        raise NotImplementedError



# @omlish-lite
from ..api import Asynclite
from .events import AsyncioAsyncliteEvents
from .identities import AsyncioAsyncliteIdentities
from .locks import AsyncioAsyncliteLocks
from .queues import AsyncioAsyncliteQueues
from .semaphores import AsyncioAsyncliteSemaphores
from .sleeps import AsyncioAsyncliteSleeps


##


class AsyncioAsynclite(
    AsyncioAsyncliteEvents,
    AsyncioAsyncliteIdentities,
    AsyncioAsyncliteLocks,
    AsyncioAsyncliteQueues,
    AsyncioAsyncliteSemaphores,
    AsyncioAsyncliteSleeps,

    Asynclite,
):
    pass



import textwrap
import typing as ta

from ... import check
from .grammars import Channel
from .grammars import Grammar
from .matches import Match
from .matches import filter_matches
from .ops import RuleRef


##


def filter_match_channels(
        m: Match,
        g: Grammar,
        *,
        keep: ta.Container[Channel] | None = None,
        remove: ta.Container[Channel] | None = None,
        keep_children: bool = False,
) -> Match:
    if keep is None and remove is None:
        return m

    def fn(x: Match) -> bool:
        if not isinstance((rr := x.op), RuleRef):
            return False

        r = check.not_none(g.rule(rr.name))

        if keep is not None and r.channel not in keep:
            return False

        if remove is not None and r.channel in remove:
            return False

        return True

    return filter_matches(
        fn,
        m,
        keep_children=keep_children,
    )


#


def only_match_rules(m: Match) -> Match:
    return filter_matches(
        lambda x: isinstance(x.op, RuleRef),
        m,
        keep_children=True,
    )


##


def fix_ws(s: str) -> str:
    return (
        textwrap.dedent(s)
        .rstrip()
        .replace('\r', '')
        .replace('\n', '\r\n')
    ) + '\r\n'


##


def get_ofs_line_column(s: str, p: int) -> tuple[int, int]:
    for i, l in enumerate(s.splitlines(keepends=True)):
        if p < (ll := len(l)):
            return i, p
        p -= ll
    raise ValueError(f'offset {p} out of range')



from .heapq import (  # noqa
    HeapqScheduledRunner,
)

# # TODO: still experimental
# from .hwt import (  # noqa
#     HwtScheduledRunner,
# )

from .tracking import (  # noqa
    TrackingScheduledRunner,
)

from .types import (  # noqa
    ScheduledFn,
    ScheduledCallback,

    ScheduledRunnerStateError,
    ScheduledCancelledError,
    ScheduledTimeoutError,

    ScheduleHandle,

    ScheduledRunnerState,
    ScheduledRunner,
)



import collections.abc
import typing as ta

from ... import check
from ... import lang
from ... import reflect as rfl
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory
from .iterables import DEFAULT_ITERABLE_CONCRETE_TYPES
from .iterables import IterableMarshaler
from .iterables import IterableUnmarshaler


##


class SequenceNotStrMarshalerFactory(MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, rfl.Protocol) and rty.cls is lang.SequenceNotStr):
            return None
        return lambda: IterableMarshaler(ctx.make_marshaler(check.single(rty.args)))


class SequenceNotStrUnmarshalerFactory(UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, rfl.Protocol) and rty.cls is lang.SequenceNotStr):
            return None
        cty = DEFAULT_ITERABLE_CONCRETE_TYPES[collections.abc.Sequence]  # type: ignore[type-abstract]
        return lambda: IterableUnmarshaler(cty, ctx.make_unmarshaler(check.single(rty.args)))



# ruff: noqa: UP006 UP045
# @omlish-lite
import abc
import types
import typing as ta

from ...lite.abstract import Abstract
from .base import AsyncliteApi
from .base import AsyncliteObject


##


class AsyncliteLock(AsyncliteObject, Abstract):
    """Non-reentrant."""

    async def __aenter__(self) -> None:
        await self.acquire()

    async def __aexit__(
            self,
            exc_type: ta.Optional[ta.Type[BaseException]] = None,
            exc_val: ta.Optional[BaseException] = None,
            exc_tb: ta.Optional[types.TracebackType] = None,
    ) -> None:
        self.release()

    @abc.abstractmethod
    def acquire(self, *, timeout: ta.Optional[float] = None) -> ta.Awaitable[None]:
        raise NotImplementedError

    @abc.abstractmethod
    def acquire_nowait(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def release(self) -> None:
        raise NotImplementedError

    class _Releaser:
        def __init__(self, lock: 'AsyncliteLock') -> None:
            self._lock = lock

        def __enter__(self) -> None:
            pass

        def __exit__(self, et, e, tb) -> None:
            self._lock.release()

    def releasing(self) -> ta.ContextManager[None]:
        return AsyncliteLock._Releaser(self)

    @abc.abstractmethod
    def locked(self) -> bool:
        raise NotImplementedError


class AsyncliteLocks(AsyncliteApi, Abstract):
    @abc.abstractmethod
    def make_lock(self) -> AsyncliteLock:
        raise NotImplementedError



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import abc
import typing as ta

from ...lite.abstract import Abstract
from ...lite.cached import cached_nullary
from ...lite.check import check


##


class YamlBackend(Abstract):
    @abc.abstractmethod
    def is_available(self) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def loads(self, s: str) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def dumps(self, o: ta.Any) -> str:
        raise NotImplementedError

    #

    def set_as_default(self) -> None:
        DEFAULT_YAML_BACKEND.INSTANCE = self


#


class NoYamlBackendAvailableError(Exception):
    pass


class FirstAvailableYamlBackend(YamlBackend):
    def __init__(self, backends: ta.Sequence[YamlBackend]) -> None:
        super().__init__()

        self._backends = backends

    def is_available(self) -> bool:
        return any(b.is_available() for b in self._backends)

    def loads(self, s: str) -> ta.Any:
        for b in self._backends:
            if b.is_available():
                return b.loads(s)
        raise NoYamlBackendAvailableError

    def dumps(self, o: ta.Any) -> str:
        for b in self._backends:
            if b.is_available():
                return b.dumps(o)
        raise NoYamlBackendAvailableError


#


class PyyamlYamlBackend(YamlBackend):
    @cached_nullary
    def _import(self) -> ta.Optional[ta.Any]:
        try:
            import yaml  # noqa
        except ImportError:
            return None
        else:
            return yaml

    def is_available(self) -> bool:
        return self._import() is not None

    def loads(self, s: str) -> ta.Any:
        return check.not_none(self._import()).safe_load(s)

    def dumps(self, o: ta.Any) -> str:
        return check.not_none(self._import()).safe_dump(o)


#


class RelativeImportGoyamlYamlBackend(YamlBackend):
    @cached_nullary
    def _import(self) -> ta.Optional[ta.Any]:
        try:
            mod = __import__('goyaml.backend', globals=globals(), level=1)
        except ImportError:
            return None
        else:
            return mod.backend

    def is_available(self) -> bool:
        return self._import() is not None

    @cached_nullary
    def _backend(self) -> YamlBackend:
        return check.not_none(self._import()).GoyamlYamlBackend()

    def loads(self, s: str) -> ta.Any:
        return self._backend().loads(s)

    def dumps(self, o: ta.Any) -> str:
        return self._backend().dumps(o)


##


@ta.final
class DEFAULT_YAML_BACKEND:  # noqa
    """This isn't just a mutable global because in amalgamated code that's not really possible."""

    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError

    INSTANCE: ta.ClassVar[YamlBackend] = FirstAvailableYamlBackend([
        PyyamlYamlBackend(),
        RelativeImportGoyamlYamlBackend(),
    ])



# ruff: noqa: UP045
# @omlish-lite
import typing as ta

from ..types import BytesLike
from .base import BaseByteStreamBufferLike
from .types import ByteStreamBufferView
from .types import MutableByteStreamBuffer


##


class ScanningByteStreamBuffer(BaseByteStreamBufferLike, MutableByteStreamBuffer):
    """
    A MutableByteStreamBuffer wrapper that caches negative-find progress to avoid repeated rescans in trickle scenarios.

    It is intentionally conservative:
      - It only caches progress for the default find range (start==0, end is None).
      - It only caches *negative* results (i.e., "-1"): once a match is found, caching is not updated, to preserve the
        property that repeated `find(sub)` on an unchanged buffer yields the same answer.

    This is designed to help framing-style code that repeatedly does:
      - buf.write(...small...)
      - buf.find(delim)
      - (not found) repeat

    Pairs well with `LongestMatchDelimiterByteStreamFrameDecoder`.
    """

    def __init__(self, buf) -> None:
        super().__init__()

        self._buf = buf
        self._scan_from_by_sub: dict[bytes, int] = {}

    @property
    def max_size(self) -> ta.Optional[int]:
        return self._buf.max_size

    #

    def __len__(self) -> int:
        return len(self._buf)

    def peek(self) -> memoryview:
        return self._buf.peek()

    def segments(self) -> ta.Sequence[memoryview]:
        return self._buf.segments()

    #

    def advance(self, n: int, /) -> None:
        self._buf.advance(n)
        self._adjust_for_consume(n)

    def split_to(self, n: int, /) -> ByteStreamBufferView:
        v = self._buf.split_to(n)
        self._adjust_for_consume(n)
        return v

    def coalesce(self, n: int, /) -> memoryview:
        return self._buf.coalesce(n)

    def find(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        if start != 0 or end is not None:
            return self._buf.find(sub, start, end)

        sub_len = len(sub)
        if sub_len <= 0:
            return self._buf.find(sub, start, end)

        scan_from = self._scan_from_by_sub.get(sub, 0)

        # Allow overlap so a match spanning old/new boundary is discoverable.
        overlap = sub_len - 1
        eff_start = scan_from - overlap
        if eff_start < 0:
            eff_start = 0

        i = self._buf.find(sub, eff_start, None)
        if i < 0:
            self._scan_from_by_sub[sub] = len(self._buf)

        return i

    def rfind(self, sub: bytes, start: int = 0, end: ta.Optional[int] = None) -> int:
        # rfind isn't the typical trickle hot-path; delegate.
        return self._buf.rfind(sub, start, end)

    #

    def write(self, data: BytesLike, /) -> None:
        self._buf.write(data)

    def reserve(self, n: int, /) -> memoryview:
        return self._buf.reserve(n)

    def commit(self, n: int, /) -> None:
        self._buf.commit(n)

    #

    def _adjust_for_consume(self, n: int) -> None:
        if not self._scan_from_by_sub:
            return

        if n <= 0:
            return

        # Only front-consumption exists in this buffer model.
        for k, v in list(self._scan_from_by_sub.items()):
            nv = v - n
            if nv <= 0:
                self._scan_from_by_sub.pop(k, None)
            else:
                self._scan_from_by_sub[k] = nv



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - logging
"""
import concurrent.futures as cf
import contextlib
import functools
import typing as ta

from ....lite.check import check
from ....sockets.addresses import SocketAndAddress
from ....sockets.bind import CanSocketBinder
from ....sockets.bind import SocketBinder
from ....sockets.server.handlers import ExecutorSocketServerHandler
from ....sockets.server.handlers import SocketHandlerSocketServerHandler
from ....sockets.server.handlers import SocketServerHandler
from ....sockets.server.handlers import SocketWrappingSocketServerHandler
from ....sockets.server.handlers import StandardSocketServerHandler
from ....sockets.server.server import SocketServer
from ....sockets.server.ssl import SslErrorHandlingSocketServerHandler
from ....sockets.server.threading import ThreadingSocketServerHandler
from ...handlers import HttpHandler
from .server import CoroHttpServer
from .sockets import CoroHttpServerSocketHandler


if ta.TYPE_CHECKING:
    import ssl


##


@contextlib.contextmanager
def make_simple_http_server(
        bind: CanSocketBinder,
        handler: HttpHandler,
        *,
        keep_alive: bool = False,
        ssl_context: ta.Optional['ssl.SSLContext'] = None,
        ignore_ssl_errors: bool = False,
        executor: ta.Optional[cf.Executor] = None,
        use_threads: bool = False,
        **kwargs: ta.Any,
) -> ta.Iterator[SocketServer]:
    check.arg(not (executor is not None and use_threads))

    #

    with contextlib.ExitStack() as es:
        server_factory = functools.partial(
            CoroHttpServer,
            handler=handler,
        )

        socket_handler = CoroHttpServerSocketHandler(
            server_factory,
            keep_alive=keep_alive,
        )

        #

        server_handler: SocketServerHandler = SocketHandlerSocketServerHandler(
            socket_handler,
        )

        #

        if ssl_context is not None:
            server_handler = SocketWrappingSocketServerHandler(
                server_handler,
                SocketAndAddress.socket_wrapper(functools.partial(
                    ssl_context.wrap_socket,
                    server_side=True,
                )),
            )

        if ignore_ssl_errors:
            server_handler = SslErrorHandlingSocketServerHandler(
                server_handler,
            )

        #

        server_handler = StandardSocketServerHandler(
            server_handler,
        )

        #

        if executor is not None:
            server_handler = ExecutorSocketServerHandler(
                server_handler,
                executor,
            )

        elif use_threads:
            server_handler = es.enter_context(ThreadingSocketServerHandler(
                server_handler,
            ))

        #

        server = es.enter_context(SocketServer(
            SocketBinder.of(bind),
            server_handler,
            **kwargs,
        ))

        yield server



from .. import lang as _lang


with _lang.auto_proxy_init(globals()):
    ##

    from .distribute import (  # noqa
        distribute_evenly,
    )

    from .prefixes import (  # noqa
        MinUniquePrefixNode,
        build_min_unique_prefix_tree,

        min_unique_prefix_lens,
        min_unique_prefix_len,
    )

    from .toposort import (  # noqa
        mut_toposort,
        toposort,
    )

    from .unify import (  # noqa
        mut_unify_sets,
        unify_sets,
    )



import contextlib
import typing as ta

from ... import lang
from .adapters import Adapter
from .columns import Columns
from .core import AsyncConn
from .core import AsyncDb
from .core import AsyncRows
from .core import AsyncTransaction
from .core import Conn
from .core import Db
from .core import Rows
from .core import Transaction
from .queries import Queryable
from .rows import Row


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


class Runner(ta.Protocol):
    def __call__(self, fn: ta.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> ta.Awaitable[T]: ...


@ta.final
class _RunnerContextManager:
    def __init__(self, runner: Runner, fn: ta.Callable, wrapper: ta.Any) -> None:
        self._runner = runner
        self._fn = fn
        self._wrapper = wrapper

    _cm: ta.Any

    async def __aenter__(self):
        self._cm = await self._runner(self._fn)
        return self._wrapper(self._runner, await self._runner(self._cm.__enter__))

    async def __aexit__(self, et, e, tb):
        return await self._runner(self._cm.__exit__, et, e, tb)


##


class _RunnerStopIteration(lang.Marker):
    pass


class SyncToAsyncRows(AsyncRows):
    def __init__(self, runner: Runner, rows: Rows) -> None:
        super().__init__()

        self._runner = runner
        self._rows = rows

    @property
    def columns(self) -> Columns:
        return self._rows.columns

    async def __anext__(self) -> Row:  # ta.Raises[StopIteration]
        def inner():
            try:
                return self._rows.__next__()
            except StopIteration:
                return _RunnerStopIteration

        if (v := await self._runner(inner)) is _RunnerStopIteration:
            raise StopAsyncIteration

        return v


class SyncToAsyncTransaction(AsyncTransaction):
    def __init__(self, runner: Runner, txn: Transaction) -> None:
        super().__init__()

        self._runner = runner
        self._txn = txn

    @property
    def adapter(self) -> Adapter:
        return self._txn.adapter

    async def commit(self) -> None:
        return await self._runner(self._txn.commit)

    async def rollback(self) -> None:
        return await self._runner(self._txn.rollback)

    def query(self, query: Queryable) -> ta.AsyncContextManager[AsyncRows]:  # ta.Raises[QueryError]
        return _RunnerContextManager(self._runner, lambda: self._txn.query(query), SyncToAsyncRows)


class SyncToAsyncConn(AsyncConn):
    def __init__(self, runner: Runner, conn: Conn) -> None:
        super().__init__()

        self._runner = runner
        self._conn = conn

    @property
    def adapter(self) -> Adapter:
        return self._conn.adapter

    def begin(self) -> ta.AsyncContextManager[AsyncTransaction]:
        return _RunnerContextManager(self._runner, self._conn.begin, SyncToAsyncTransaction)

    def query(self, query: Queryable) -> ta.AsyncContextManager[AsyncRows]:  # ta.Raises[QueryError]
        return _RunnerContextManager(self._runner, lambda: self._conn.query(query), SyncToAsyncRows)


class SyncToAsyncDb(AsyncDb):
    def __init__(
            self,
            runner_factory: ta.Callable[[], ta.AsyncContextManager[Runner]],
            db: Db,
    ) -> None:
        super().__init__()

        self._runner_factory = runner_factory
        self._db = db

    @property
    def adapter(self) -> Adapter:
        return self._db.adapter

    async def _connect(self, aes: contextlib.AsyncExitStack) -> AsyncConn:
        runner = await aes.enter_async_context(self._runner_factory())

        rcm = _RunnerContextManager(runner, self._db.connect, SyncToAsyncConn)
        return await aes.enter_async_context(rcm)

    def connect(self) -> ta.AsyncContextManager[AsyncConn]:
        @contextlib.asynccontextmanager
        async def inner():
            async with contextlib.AsyncExitStack() as aes:
                yield await self._connect(aes)

        return inner()

    def query(self, query: Queryable) -> ta.AsyncContextManager[AsyncRows]:  # ta.Raises[QueryError]
        @contextlib.asynccontextmanager
        async def inner():
            async with contextlib.AsyncExitStack() as aes:
                conn = await self._connect(aes)
                yield await aes.enter_async_context(conn.query(query))

        return inner()



import collections.abc
import typing as ta

from ... import dataclasses as dc
from ... import lang
from ... import marshal as msh
from .base import Node
from .exprs import CanExpr
from .exprs import Expr
from .exprs import ExprBuilder
from .idents import CanIdent
from .idents import Ident
from .relations import CanRelation
from .relations import Relation
from .relations import RelationBuilder
from .stmts import Stmt


##


class Field(Node, lang.Final):
    c: Ident
    v: Expr


class Fields(Node, lang.Final):
    fields: ta.Sequence[Field] = dc.xfield(coerce=tuple)


class Update(Stmt, lang.Final):
    into: Relation
    fields: Fields
    # from_: Relation = dc.xfield(repr_fn=lang.opt_repr) | msh.with_field_options(name='from', omit_if=lang.is_none)
    where: Expr | None = dc.xfield(None, repr_fn=lang.opt_repr) | msh.with_field_options(omit_if=lang.is_none)


CanField: ta.TypeAlias = Field | tuple[CanIdent, CanExpr]

CanFields: ta.TypeAlias = ta.Union[  # noqa
    Fields,
    ta.Mapping[CanIdent, CanExpr],
    ta.Sequence[CanField],
]


class UpdateBuilder(RelationBuilder, ExprBuilder):
    def field(self, f: CanField) -> Field:
        if isinstance(f, Field):
            return f
        elif isinstance(f, tuple):
            c, v = f
            return Field(
                self.ident(c),
                self.expr(v),
            )
        else:
            raise TypeError(f)

    def fields(self, fs: CanFields) -> Fields:
        if isinstance(fs, Fields):
            return fs
        elif isinstance(fs, collections.abc.Mapping):
            return Fields(tuple(self.field(i) for i in fs.items()))
        elif isinstance(fs, collections.abc.Sequence):
            return Fields(tuple(self.field(i) for i in fs))
        else:
            raise TypeError(fs)

    def update(
            self,
            into: CanRelation,
            fields: Fields | CanFields,
            *,
            # from_: CanRelation | None = None,
            where: CanExpr | None = None,
    ) -> Update:
        return Update(
            into=self.relation(into),
            fields=self.fields(fields),
            where=self.expr(where) if where is not None else None,
        )



# ruff: noqa: UP006 UP045
# @omlish-lite
import collections
import dataclasses as dc
import typing as ta

from ....io.pipelines.bytes.buffering import InboundBytesBufferingIoPipelineHandler
from ....io.pipelines.core import IoPipelineHandlerContext
from ....io.pipelines.core import IoPipelineMessages
from ....io.pipelines.flow.types import IoPipelineFlow
from ....io.pipelines.flow.types import IoPipelineFlowMessages
from ....io.streams.utils import ByteStreamBuffers
from ....io.types import BytesLike
from ....lite.abstract import Abstract
from ..objects import IoPipelineHttpMessageBodyData
from ..objects import IoPipelineHttpMessageEnd
from ..objects import IoPipelineHttpMessageHead
from ..objects import IoPipelineHttpMessageObjects
from .codings import DefaultIoPiplineHttpCompressionCodings
from .codings import IoPiplineHttpDecompressorCoding
from .codings import IoPiplineHttpDecompressorCodings


##


@dc.dataclass(frozen=True)
class IoPipelineHttpDecompressionConfig:
    DEFAULT: ta.ClassVar['IoPipelineHttpDecompressionConfig']

    max_decomp_chunk: int = 64 * 1024  # max bytes emitted per inflate step

    max_decomp_total: ta.Optional[int] = None    # max total decompressed bytes per object
    max_expansion_ratio: ta.Optional[int] = 200  # max_out <= max(1, in_total) * ratio (+ small slack)

    max_out_pending: ta.Optional[int] = 256 * 1024  # cap decompressed bytes retained by this stage (if you buffer)

    # CPU Bounding: how many decompress steps to perform before yielding to the driver
    max_steps_per_call: ta.Optional[int] = None


IoPipelineHttpDecompressionConfig.DEFAULT = IoPipelineHttpDecompressionConfig()


#


class IoPipelineHttpObjectDecompressor(
    IoPipelineHttpMessageObjects,
    InboundBytesBufferingIoPipelineHandler,
    Abstract,
):
    def __init__(
            self,
            codings: ta.Optional[IoPiplineHttpDecompressorCodings] = None,
            config: IoPipelineHttpDecompressionConfig = IoPipelineHttpDecompressionConfig.DEFAULT,
    ) -> None:
        super().__init__()

        self._config = config
        if codings is None:
            codings = DefaultIoPiplineHttpCompressionCodings.DECOMPRESSOR
        self._codings = codings

        self._decompressor: ta.Optional[IoPiplineHttpDecompressorCoding] = None

        # Statistics for budget checks
        self._in_total_bytes = 0
        self._out_total_bytes = 0

        # Internal buffering
        self._in_pending: collections.deque[BytesLike] = collections.deque()
        self._in_pending_bytes = 0
        self._out_pending: collections.deque[BytesLike] = collections.deque()
        self._out_pending_bytes = 0

        # Flow Control and Deferral State
        self._read_requested = False
        self._pending_end: ta.Optional[IoPipelineHttpMessageEnd] = None

    #

    def inbound_buffered_bytes(self) -> int:
        return self._in_pending_bytes + self._out_pending_bytes

    #

    def _reset(self) -> None:
        self._decompressor = None

        self._in_total_bytes = 0
        self._out_total_bytes = 0

        self._in_pending.clear()
        self._in_pending_bytes = 0
        self._out_pending.clear()
        self._out_pending_bytes = 0

        self._pending_end = None

    def _check_budgets(self) -> None:
        if (mdt := self._config.max_decomp_total) is not None and self._out_total_bytes > mdt:
            raise ValueError('decompressor output exceeds limit (possible zip bomb)')

        if (mer := self._config.max_expansion_ratio) is not None:
            slack = self._config.max_decomp_chunk
            if self._out_total_bytes > (max(1, self._in_total_bytes) * mer + slack):
                raise ValueError('decompressor expansion ratio exceeds limit (possible zip bomb)')

    def _is_auto_read(self, ctx: IoPipelineHandlerContext) -> bool:
        if (flow := ctx.services.find(IoPipelineFlow)) is None:
            return True
        return flow.is_auto_read()

    def _emit_out_pending(self, ctx: IoPipelineHandlerContext) -> bool:
        """Returns True if at least one message was emitted."""

        emitted = False

        while self._out_pending and (self._is_auto_read(ctx) or self._read_requested):
            o = self._out_pending.popleft()
            self._out_pending_bytes -= len(o)

            if not self._is_auto_read(ctx):
                self._read_requested = False

            ctx.feed_in(self._make_body_data(o))
            emitted = True

            # In manual mode, we satisfy one 'read' at a time.
            if not self._is_auto_read(ctx):
                break

        return emitted

    def _pump(self, ctx: IoPipelineHandlerContext) -> bool:
        """Returns True if it effectively satisfied a read request."""

        z = self._decompressor
        if z is None:
            return False

        steps = 0
        max_steps = self._config.max_steps_per_call

        # 1. Try to clear existing output.
        if self._emit_out_pending(ctx):
            if not self._is_auto_read(ctx):
                return True

        # 2. If blocked by downstream, we can't satisfy anything.
        if self._out_pending:
            return False

        # 3. Decompression Loop
        while self._in_pending:
            # Enforce output buffer budget
            if (mop := self._config.max_out_pending) is not None:
                if self._out_pending_bytes >= mop:
                    break

            # Check for CPU step limit
            if max_steps is not None and steps >= max_steps:
                self._defer_resume(ctx)
                return False  # We haven't satisfied it yet, we deferred.

            steps += 1
            chunk = self._in_pending.popleft()
            cl = len(chunk)
            self._in_pending_bytes -= cl

            out = z.decompress(chunk, self._config.max_decomp_chunk)
            if out:
                ol = len(out)
                self._out_total_bytes += ol
                self._out_pending.append(out)
                self._out_pending_bytes += ol
                self._check_budgets()

                if self._emit_out_pending(ctx):
                    if not self._is_auto_read(ctx):
                        return True  # Satisfied!

            ut = z.unconsumed_tail()
            if ut:
                self._in_pending.appendleft(ut)
                self._in_pending_bytes += len(ut)
                if not out:
                    break

        # 4. Handle EOF
        if not self._in_pending and self._pending_end is not None:
            if max_steps is not None and steps >= max_steps:
                self._defer_resume(ctx)
                return False

            out = z.finish()
            if out:
                ol = len(out)
                self._out_total_bytes += ol
                self._out_pending.append(out)
                self._out_pending_bytes += ol
                self._check_budgets()
                self._emit_out_pending(ctx)

            msg = self._pending_end
            self._pending_end = None
            self._read_requested = False
            ctx.feed_in(msg)
            return True  # FinalInput counts as satisfying the last read

        return False

    def _defer_resume(self, ctx: IoPipelineHandlerContext) -> None:
        def resume(c: IoPipelineHandlerContext) -> None:
            # If a deferred pump satisfies a read, it must provide the FlushInput
            if self._pump(c) and not self._is_auto_read(c):
                c.feed_in(IoPipelineFlowMessages.FlushInput())

        ctx.defer(resume)

    #

    def _on_inbound_final_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineMessages.FinalInput) -> None:
        if self._decompressor is None:
            ctx.feed_in(msg)
            return

        self._reset()

        ctx.feed_in(self._make_aborted('eof before end of message'))
        ctx.feed_in(msg)

    def _on_inbound_flush_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineFlowMessages.FlushInput) -> None:
        self._pump(ctx)
        ctx.feed_in(msg)

    def _on_inbound_head(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageHead) -> None:
        if self._decompressor is not None:
            ctx.feed_in(self._make_aborted('unexpected message sequence'))
            return

        enc = msg.headers.lower.get('content-encoding', ())

        # TODO: spec is actually an ordered stack lol
        for coding_name, coding in self._codings.items():
            if coding_name.lower() in enc:
                self._decompressor = coding()
                break

        ctx.feed_in(msg)

    def _on_inbound_body_data(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageBodyData) -> None:
        if self._decompressor is None:
            ctx.feed_in(msg)
            return

        for mv in ByteStreamBuffers.iter_segments(msg.data):
            mvl = len(mv)
            self._in_total_bytes += mvl
            self._in_pending.append(mv)
            self._in_pending_bytes += mvl
            self._check_budgets()

        self._pump(ctx)

    def _on_inbound_end(self, ctx: IoPipelineHandlerContext, msg: IoPipelineHttpMessageEnd) -> None:
        if self._decompressor is None:
            ctx.feed_in(msg)
            return

        self._pending_end = msg
        self._pump(ctx)

    def inbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineMessages.FinalInput):
            self._on_inbound_final_input(ctx, msg)

        elif isinstance(msg, IoPipelineFlowMessages.FlushInput):
            self._on_inbound_flush_input(ctx, msg)

        elif isinstance(msg, self._head_type):
            self._on_inbound_head(ctx, msg)

        elif isinstance(msg, self._body_data_type):
            self._on_inbound_body_data(ctx, msg)

        elif isinstance(msg, self._end_type):
            self._on_inbound_end(ctx, msg)

        else:
            ctx.feed_in(msg)

    #

    def _on_outbound_ready_for_input(self, ctx: IoPipelineHandlerContext, msg: IoPipelineFlowMessages.ReadyForInput) -> None:  # Noqa
        self._read_requested = True

        if self._out_pending or (self._decompressor is not None and self._in_pending):
            if self._pump(ctx):
                if not self._is_auto_read(ctx):
                    ctx.feed_in(IoPipelineFlowMessages.FlushInput())

                return  # Swallow since we satisfied it

        ctx.feed_out(msg)

    def outbound(self, ctx: IoPipelineHandlerContext, msg: ta.Any) -> None:
        if isinstance(msg, IoPipelineFlowMessages.ReadyForInput):
            self._on_outbound_ready_for_input(ctx, msg)

        else:
            ctx.feed_out(msg)



import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import reflect as rfl
from ..api.contexts import MarshalContext
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory
from ..api.values import Value


##


@dc.dataclass(frozen=True)
class OptionalMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Any | None) -> Value:
        if o is None:
            return None
        return self.e.marshal(ctx, o)


class OptionalMarshalerFactory(MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, rfl.Union) and rty.is_optional):
            return None
        return lambda: OptionalMarshaler(ctx.make_marshaler(check.isinstance(rty, rfl.Union).without_none()))


@dc.dataclass(frozen=True)
class OptionalUnmarshaler(Unmarshaler):
    e: Unmarshaler

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any | None:
        if v is None:
            return None
        return self.e.unmarshal(ctx, v)


class OptionalUnmarshalerFactory(UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, rfl.Union) and rty.is_optional):
            return None
        return lambda: OptionalUnmarshaler(ctx.make_unmarshaler(check.isinstance(rty, rfl.Union).without_none()))



"""
See:
- https://stackoverflow.com/a/75903610
- https://newosxbook.com/bonus/vol1ch16.html
"""
import ctypes as ct
import dataclasses as dc
import errno
import socket
import struct
import sys
import typing as ta

from .. import lang


##


@dc.dataclass(frozen=True)
class PeerCred:
    pid: int | None = None
    uid: int | None = None
    gid: int | None = None

    def __repr__(self) -> str:
        return ''.join([
            f'{self.__class__.__name__}(',
            ', '.join([f'{f}={v}' for f in ('pid', 'uid', 'gid') if (v := getattr(self, f)) is not None]),
            ')',
        ])


#


def _get_linux_socket_peer_cred(sock: socket.socket) -> PeerCred | None:
    # Linux: SO_PEERCRED => struct ucred { pid_t pid; uid_t uid; gid_t gid; }

    if not hasattr(socket, 'SO_PEERCRED'):
        return None

    try:
        raw = sock.getsockopt(socket.SOL_SOCKET, socket.SO_PEERCRED, struct.calcsize('3i'))
    except OSError as e:
        if e.errno == errno.ENOTCONN:
            return None
        raise

    pid, uid, gid = struct.unpack('3i', raw)
    return PeerCred(
        pid=int(pid),
        uid=int(uid),
        gid=int(gid),
    )


#


@lang.cached_function
def _get_bsd_libc() -> ta.Any:
    libc = ct.CDLL(None, use_errno=True)

    if hasattr(libc, 'getpeereid'):
        libc.getpeereid.argtypes = [
            ct.c_int,
            ct.POINTER(ct.c_uint),
            ct.POINTER(ct.c_uint),
        ]
        libc.getpeereid.restype = ct.c_int

    return libc


class _BsdConsts(lang.Namespace):
    SOL_LOCAL = 1

    LOCAL_PEERCRED = 1
    LOCAL_PEERPID = 2
    LOCAL_PEEREPID = 3
    LOCAL_PEERUUID = 4
    LOCAL_PEEREUUID = 5
    LOCAL_PEERTOKEN = 6


def _get_bsd_socket_peer_cred(sock: socket.socket) -> PeerCred | None:
    # Darwin / BSD-ish path:
    # - try LOCAL_PEERPID for pid if exposed by Python's socket module
    # - try getpeereid() for euid/egid

    pid: int | None = None
    uid: int | None = None
    gid: int | None = None

    try:
        raw = sock.getsockopt(_BsdConsts.SOL_LOCAL, _BsdConsts.LOCAL_PEERPID, struct.calcsize('i'))
    except OSError as e:
        if e.errno == errno.ENOTCONN:
            return None
        raise

    [pid] = struct.unpack('i', raw)  # noqa

    libc = _get_bsd_libc()

    if hasattr(libc, 'getpeereid'):
        euid = ct.c_uint()
        egid = ct.c_uint()

        if rc := libc.getpeereid(sock.fileno(), ct.byref(euid), ct.byref(egid)):  # noqa
            if (err := ct.get_errno()) == errno.ENOTCONN:
                return None
            raise OSError(err, 'ctypes call failed')

        uid = int(euid.value)
        gid = int(egid.value)

    return PeerCred(
        pid=pid,
        uid=uid,
        gid=gid,
    )


#


def get_unix_socket_peer_cred(sock: socket.socket) -> PeerCred | None:
    if sock.family != socket.AF_UNIX:
        return None

    if sock.fileno() < 0:
        return None

    sys_platform = getattr(sys, 'platform')  # shuts up mypy

    # Linux: SO_PEERCRED => struct ucred { pid_t pid; uid_t uid; gid_t gid; }
    if sys_platform.startswith('linux'):
        return _get_linux_socket_peer_cred(sock)

    if (
            sys_platform == 'darwin' or
            sys_platform.startswith(('freebsd', 'openbsd', 'netbsd'))
    ):
        return _get_bsd_socket_peer_cred(sock)

    return None



import dataclasses as dc
import typing as ta

from .. import lang


T = ta.TypeVar('T')


##


@dc.dataclass()
class UniqueStats:
    key: ta.Any
    num_seen: int
    first_idx: int
    last_idx: int


@dc.dataclass(frozen=True)
class UniqueItem(ta.Generic[T]):
    idx: int
    item: T
    stats: UniqueStats
    out: lang.Maybe[T]


class UniqueIterator(ta.Iterator[UniqueItem[T]]):
    def __init__(
            self,
            it: ta.Iterable[T],
            keyer: ta.Callable[[T], ta.Any] = lang.identity,
    ) -> None:
        super().__init__()

        self._it = enumerate(it)
        self._keyer = keyer

        self.stats: dict[ta.Any, UniqueStats] = {}

    def __next__(self) -> UniqueItem[T]:
        idx, item = next(self._it)
        key = self._keyer(item)

        try:
            stats = self.stats[key]

        except KeyError:
            stats = self.stats[key] = UniqueStats(
                key,
                num_seen=1,
                first_idx=idx,
                last_idx=idx,
            )
            return UniqueItem(
                idx,
                item,
                stats,
                lang.just(item),
            )

        else:
            stats.num_seen += 1
            stats.last_idx = idx
            return UniqueItem(
                idx,
                item,
                stats,
                lang.empty(),
            )



import abc
import threading

from ... import check
from ... import lang
from .types import ScheduledRunner
from .types import ScheduledRunnerState
from .types import ScheduledRunnerStateError


##


class BaseScheduledRunner(ScheduledRunner, lang.Abstract):
    def __init__(self) -> None:
        super().__init__()

        self._cv = threading.Condition()

    #

    _state: ScheduledRunnerState = ScheduledRunnerState.NEW

    def get_state(self) -> ScheduledRunnerState:
        with self._cv:
            return self._state

    def _check_state(self, *valid_states: ScheduledRunnerState) -> ScheduledRunnerState:
        if self._state not in valid_states:
            raise ScheduledRunnerStateError(self._state)
        return self._state

    def _check_state_not(self, *invalid_states: ScheduledRunnerState) -> ScheduledRunnerState:
        if self._state in invalid_states:
            raise ScheduledRunnerStateError(self._state)
        return self._state

    #

    _thread: threading.Thread | None = None

    def shutdown(
            self,
            *,
            cancel_all: bool = False,
            no_wait: bool = False,
            timeout: float | None = None,
    ) -> None:
        with self._cv:
            if self._state == ScheduledRunnerState.NEW:
                check.none(self._thread)
                return

            elif self._state == ScheduledRunnerState.RUNNING:
                if cancel_all:
                    self._state = ScheduledRunnerState.STOPPING
                else:
                    self._state = ScheduledRunnerState.DRAINING
                self._cv.notify()

            elif self._state in (ScheduledRunnerState.STOPPING, ScheduledRunnerState.DRAINING):
                pass

            elif self._state == ScheduledRunnerState.STOPPED:
                return

            else:
                raise ScheduledRunnerStateError(self._state)

        if no_wait:
            return

        self.wait(timeout=timeout)

    def wait(self, timeout: float | None = None) -> None:
        with self._cv:
            if self._state == ScheduledRunnerState.STOPPED:
                return

            self._check_state(ScheduledRunnerState.STOPPING, ScheduledRunnerState.DRAINING)

        if self._thread is not None and self._thread.is_alive():
            self._thread.join(timeout=timeout)

            if self._thread.is_alive():
                raise TimeoutError

        with self._cv:
            self._check_state(ScheduledRunnerState.STOPPED)

    #

    def _ensure_thread_locked(self) -> None:
        if self._thread is not None:
            self._check_state_not(ScheduledRunnerState.NEW)
            return

        self._check_state(ScheduledRunnerState.NEW)

        t = threading.Thread(
            target=self.__thread_main,
            name=type(self).__name__,
            daemon=True,
        )

        self._thread = t

        t.start()

        self._state = ScheduledRunnerState.RUNNING

    def __thread_main(self) -> None:
        with self._cv:
            # May have requested shutdown before thread got scheduled
            if self._state in (ScheduledRunnerState.STOPPING, ScheduledRunnerState.DRAINING):
                self._state = ScheduledRunnerState.STOPPED
                return

            self._check_state(ScheduledRunnerState.RUNNING)
            check.state(self._thread is threading.current_thread())

        try:
            self._thread_main()

        finally:
            with self._cv:
                check.state(self._thread is threading.current_thread())
                self._state = ScheduledRunnerState.STOPPED

    @abc.abstractmethod
    def _thread_main(self) -> None:
        raise NotImplementedError



import asyncio
import concurrent.futures as cf
import contextvars
import functools
import typing as ta


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


@ta.final
class AsyncioToThread:
    def __init__(
            self,
            loop: asyncio.AbstractEventLoop | None = None,
            exe: cf.Executor | None = None,
    ) -> None:
        self._loop = loop
        self._exe = exe

    async def __call__(self, func: ta.Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        if (loop := self._loop) is None:
            loop = asyncio.get_running_loop()
        ctx = contextvars.copy_context()
        func_call = functools.partial(ctx.run, func, *args, **kwargs)
        return await loop.run_in_executor(self._exe, func_call)



import threading
import typing as ta

from ... import check
from ... import reflect as rfl
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory


FactoryT = ta.TypeVar('FactoryT', bound=MarshalerFactory | UnmarshalerFactory)


##


class _TypeCacheFactory(ta.Generic[FactoryT]):
    def __init__(self, fac: FactoryT) -> None:
        super().__init__()

        self._fac = fac

        self._dct: dict[rfl.Type, ta.Any | None] = {}
        self._lock = threading.RLock()

    def _make(self, rty, dfl):
        check.isinstance(rty, rfl.TYPES)

        try:
            return self._dct[rty]
        except KeyError:
            pass

        with self._lock:
            try:
                return self._dct[rty]
            except KeyError:
                pass

            if (m := dfl()) is None:
                self._dct[rty] = None
                return None

            x = None

            def inner():
                nonlocal x
                if x is None:
                    x = m()
                return x

            self._dct[rty] = inner
            return inner


class TypeCacheMarshalerFactory(_TypeCacheFactory[MarshalerFactory], MarshalerFactory):
    def make_marshaler(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        return self._make(rty, lambda: self._fac.make_marshaler(ctx, rty))


class TypeCacheUnmarshalerFactory(_TypeCacheFactory[UnmarshalerFactory], UnmarshalerFactory):
    def make_unmarshaler(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        return self._make(rty, lambda: self._fac.make_unmarshaler(ctx, rty))



import contextlib
import typing as ta

from .. import check
from .. import collections as col
from .. import dataclasses as dc
from .. import inject as inj
from .. import lang
from .base import AsyncLifecycle
from .base import Lifecycle
from .contextmanagers import async_lifecycle_context_manage
from .contextmanagers import lifecycle_context_manage
from .manager import AsyncLifecycleManager
from .manager import LifecycleManager
from .unwrap import unwrap_any_lifecycle


##


@ta.final
class _LifecycleRegistrar(lang.Final):
    def __init__(
            self,
            lifecycle_manager_cls: type[LifecycleManager | AsyncLifecycleManager],
            lifecycle_cls_tup: tuple[type[Lifecycle | AsyncLifecycle], ...],
            *,
            unwrap: bool = False,
    ) -> None:
        super().__init__()

        self._lifecycle_manager_cls = lifecycle_manager_cls
        self._lifecycle_cls_tup = lifecycle_cls_tup
        self._unwrap = unwrap

        self._seen: ta.MutableSet[ta.Any] = col.IdentityWeakSet()
        self._stack: list[_LifecycleRegistrar.State] = []

    @ta.final
    @dc.dataclass(frozen=True)
    class Dep(lang.Final):
        binding: inj.Binding | None
        obj: ta.Any
        lco: ta.Any

    @ta.final
    @dc.dataclass(frozen=True)
    class State(lang.Final):
        key: inj.Key
        deps: list['_LifecycleRegistrar.Dep'] = dc.field(default_factory=list)

    async def _on_provision(
            self,
            injector: inj.AsyncInjector,
            key: inj.Key,
            binding: inj.Binding | None,
            fn: ta.Callable[[], ta.Awaitable[ta.Any]],
    ) -> ta.Awaitable[ta.Any]:
        st = _LifecycleRegistrar.State(key)
        self._stack.append(st)
        try:
            obj = await fn()
        finally:
            popped = self._stack.pop()
            check.state(popped is st)

        lco: ta.Any
        if self._unwrap:
            lco = unwrap_any_lifecycle(obj)
        else:
            lco = obj

        if (
                lco is not None and
                isinstance(lco, self._lifecycle_cls_tup) and
                not isinstance(obj, self._lifecycle_manager_cls)
        ):
            if self._stack:
                self._stack[-1].deps.append(_LifecycleRegistrar.Dep(binding, obj, lco))

            if obj not in self._seen:
                mgr = await injector[self._lifecycle_manager_cls]

                dep_lcos = [d.lco for d in st.deps]

                if isinstance(mgr, AsyncLifecycleManager):
                    await mgr.add(lco, dep_lcos)
                elif isinstance(mgr, LifecycleManager):
                    mgr.add(lco, dep_lcos)
                else:
                    raise TypeError(mgr)

                self._seen.add(obj)

        elif self._stack:
            self._stack[-1].deps.extend(st.deps)

        return obj


def bind_lifecycle_registrar() -> inj.Elements:
    return inj.as_elements(
        inj.bind(_LifecycleRegistrar, to_const=(lr := _LifecycleRegistrar(
            LifecycleManager,
            (Lifecycle,),
            unwrap=True,
        ))),
        inj.bind_provision_listener(lr._on_provision),  # noqa
    )


def bind_async_lifecycle_registrar() -> inj.Elements:
    return inj.as_elements(
        inj.bind(_LifecycleRegistrar, to_const=(lr := _LifecycleRegistrar(
            AsyncLifecycleManager,
            (Lifecycle, AsyncLifecycle),
            unwrap=True,
        ))),
        inj.bind_provision_listener(lr._on_provision),  # noqa
    )


##


def bind_managed_lifecycle_manager(*, eager: bool | int = False) -> inj.Elements:
    # FIXME: lock?
    def inner(es: contextlib.ExitStack) -> LifecycleManager:
        return es.enter_context(lifecycle_context_manage(LifecycleManager()))

    return inj.as_elements(
        inj.bind(inner, singleton=True, eager=eager),
    )


def bind_async_managed_lifecycle_manager(*, eager: bool | int = False) -> inj.Elements:
    # FIXME: lock?
    async def inner(aes: contextlib.AsyncExitStack) -> AsyncLifecycleManager:
        return await aes.enter_async_context(async_lifecycle_context_manage(AsyncLifecycleManager()))

    return inj.as_elements(
        inj.bind(inner, singleton=True, eager=eager),
    )



# ruff: noqa: UP007
"""
TODO:
 - cext _init_typed_values_collection
"""
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .accessor import TypedValuesAccessor
from .consumer import TypedValuesConsumer
from .values import TypedValue
from .values import UniqueTypedValue


TypedValueT = ta.TypeVar('TypedValueT', bound='TypedValue')
TypedValueU = ta.TypeVar('TypedValueU', bound='TypedValue')

UniqueTypedValueT = ta.TypeVar('UniqueTypedValueT', bound='UniqueTypedValue')


##


@dc.dataclass()
class DuplicateUniqueTypedValueError(Exception):
    cls: type
    new: TypedValue
    old: TypedValue


##


def _init_typed_values_collection(
        *tvs: TypedValueT,
        override: bool = False,
        check_type: type | tuple[type, ...] | None = None,
) -> tuple[
    tuple[TypedValueT, ...],
    dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]],
    dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]],
]:
    if not tvs:
        return ((), {}, {})

    # Either a non-unique TypedValue or a tuple of the form (unique_tv_cls, tv, unique_lst, idx_in_unique_lst). Notably,
    # this intermediate list has the 'opposite' form of the returned collections: where the output dicts have a scalar
    # tv for unique types and a sequence of tv's for non-unique types, this has scalar values for non-unique types and a
    # tuple (heterogeneous, however) for unique types.
    tmp_lst: list[ta.Union[
        TypedValueT,
        tuple[
            type,
            TypedValueT,
            list[TypedValueT],
            int,
        ],
    ]] = []

    # When override is False duplicate unique values raises early. When override is True, however, last-in-wins. This
    # could probably rely on dict insertion order preservation and just overwrite in-place, but it's intentionally done
    # explicitly: preservation of tv ordering in all aspects is crucial, and retention of some intermediates eases
    # debugging and error reporting.
    unique_dct: dict[type, list[TypedValueT]] = {}

    for tv in tvs:
        if check_type is not None:
            if not isinstance(tv, check_type):
                raise TypeError(tv)

        if isinstance(tv, UniqueTypedValue):
            unique_tv_cls = tv._unique_typed_value_cls  # noqa

            if not override:
                try:
                    exu = unique_dct[unique_tv_cls]
                except KeyError:
                    pass
                else:
                    raise DuplicateUniqueTypedValueError(unique_tv_cls, tv, exu[0])

            unique_lst = unique_dct.setdefault(unique_tv_cls, [])
            unique_lst.append(tv)

            tmp_lst.append((unique_tv_cls, tv, unique_lst, len(unique_lst)))

        elif isinstance(tv, TypedValue):
            tmp_lst.append(tv)

        else:
            raise TypeError(tv)

    # The output list with input order preserved and absent of overridden uniques.
    lst: list[TypedValueT] = []

    # This dict has the expected form: scalar tv's for unique types, and an accumulating list for non-unique types.
    tmp_dct: dict[type, TypedValueT | list[TypedValueT]] = {}

    for obj in tmp_lst:
        # Unique type
        if isinstance(obj, tuple):
            unique_tv_cls, tv, unique_lst, idx = obj

            # Last-in-wins
            if idx == len(unique_lst):
                lst.append(tv)
                tmp_dct[unique_tv_cls] = tv

        else:
            tv = obj
            lst.append(tv)
            tmp_dct.setdefault(type(tv), []).append(tv)  # type: ignore[union-attr]

    # This is the 'canonical' output dict: scalar tv's for unique types keyed by their unique type, and homogenous
    # tuples of tv's keyed by their instance type for non-unique types.
    dct: dict[type, TypedValueT | tuple[TypedValueT, ...]] = {
        k: tuple(v) if isinstance(v, list) else v
        for k, v in tmp_dct.items()
    }

    # This is the secondary output dict: the contents of previous dict in addition to entries of unique tv's keyed by
    # their instance type. Notably, for unique tv's in which their unique type *is* their instance type (which is
    # perfectly fine) this will squash together duplicate (k, v) pairs, which is also perfectly fine.
    dct2: dict[type, TypedValueT | tuple[TypedValueT, ...]] = {
        **dct,
        **{type(v): v for v in dct.values() if isinstance(v, UniqueTypedValue)},
    }

    return (tuple(lst), dct, dct2)


##


try:
    from . import _collection  # type: ignore
except ImportError:
    pass
else:
    _init_typed_values_collection = _collection.init_typed_values_collection  # noqa


##


@ta.final
class TypedValues(
    TypedValuesAccessor[TypedValueT],
    lang.Final,
    ta.Generic[TypedValueT],
):
    def __init__(
            self,
            *tvs: TypedValueT,
            override: bool = False,
            check_type: type | tuple[type, ...] | None = None,
    ) -> None:
        self._tup, self._dct, self._dct2 = _init_typed_values_collection(*tvs, override=override, check_type=check_type)  # noqa

    _tup: tuple[TypedValueT, ...]

    # For non unique types, a map from tv instance type to a tuple of instances of that type. For unique tv types, a map
    # from tv unique type to the tv for that unique tv type.
    _dct: dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]

    # The contents of the previous dict in addition to entries from unique tv's keyed by their instance type.
    _dct2: dict[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]

    #

    @property
    def debug(self) -> ta.Sequence[TypedValueT]:
        return self._tup

    #

    def without(self, *tys: type) -> ta.Iterator[TypedValueT]:
        for o in self._tup:
            if tys and isinstance(o, tys):
                continue
            yield o

    #

    def discard(self, *tys: type) -> 'TypedValues':
        nl = list(self.without(*tys))

        if len(nl) == len(self._tup):
            return self

        return TypedValues(*self.without(*tys))

    def update(
            self,
            *tvs,
            discard: ta.Iterable[type] | None = None,
            override: bool = False,
    ) -> 'TypedValues':
        if not tvs and not discard:
            return self

        n = TypedValues(
            *(self.discard(*discard) if discard else self._tup),
            *tvs,
            override=override,
        )

        if lang.seqs_identical(self._tup, n._tup):
            return self

        return n

    #

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(map(repr, self._tup))})'

    #

    _set: frozenset[TypedValueT]

    def to_set(self) -> frozenset[TypedValueT]:
        try:
            return self._set
        except AttributeError:
            pass
        s = frozenset(self._tup)
        self._set = s
        return s

    #

    _hash: int

    def __hash__(self) -> int:
        try:
            return self._hash
        except AttributeError:
            pass
        h = hash(self._tup)
        self._hash = h
        return h

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return (
            other is self or
            other._tup == self._tup
        )

    #

    def __iter__(self) -> ta.Iterator[TypedValueT]:
        return iter(self._tup)

    def __len__(self) -> int:
        return len(self._tup)

    def __bool__(self) -> bool:
        return bool(self._tup)

    #

    def keys(self) -> ta.KeysView[type[TypedValueT]]:
        return self._dct.keys()

    def values(self) -> ta.ValuesView[TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.values()

    def items(self) -> ta.ItemsView[type[TypedValueT], TypedValueT | tuple[TypedValueT, ...]]:
        return self._dct.items()

    #

    def consume(self) -> TypedValuesConsumer[TypedValueT]:
        return TypedValuesConsumer(self._dct.items())

    #

    def _typed_value_contains(self, cls):
        return cls in self._dct2

    def _typed_value_getitem(self, key):
        if isinstance(key, int):
            return self._tup[key]
        elif isinstance(key, type):
            return self._dct2[check.issubclass(key, TypedValue)]
        else:
            raise TypeError(key)

    _any_dct: dict[type | tuple[type, ...], tuple[TypedValueT, ...]]

    def _typed_value_get_any(self, cls):
        try:
            any_dct = self._any_dct
        except AttributeError:
            any_dct = {}
            self._any_dct = any_dct

        try:
            return any_dct[cls]
        except KeyError:
            pass

        ret = tuple(tv for tv in self if isinstance(tv, cls))
        any_dct[cls] = ret
        return ret


collect = TypedValues


def as_collection(src: ta.Sequence[TypedValueT]) -> TypedValues[TypedValueT]:
    if isinstance(src, TypedValues):
        return src
    else:
        return TypedValues(*src)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
"""
TODO:
 - ?? lite ?? whole point is an async client in lite code
 - stream
  - chunk size - httpx interface is awful, punch through?
 - httpx catch
 - return non-200 HttpResponses
 - async
"""
import dataclasses as dc
import typing as ta

from ...lite.abstract import Abstract
from ...lite.attrops import AttrOps
from ...lite.cached import cached_property
from ..headers import CanHttpHeaders
from ..headers import HttpHeaders


BaseHttpClientResponseT = ta.TypeVar('BaseHttpClientResponseT', bound='BaseHttpClientResponse')


##


DEFAULT_HTTP_CLIENT_ENCODING = 'utf-8'


def is_success_http_status(status: int) -> bool:
    return 200 <= status < 300


##


@ta.final
@dc.dataclass(frozen=True)
class HttpClientRequest:
    url: str
    method: ta.Optional[str] = None  # noqa

    # _: dc.KW_ONLY

    headers: ta.Optional[CanHttpHeaders] = None
    data: ta.Union[bytes, str, None] = None

    timeout_s: ta.Optional[float] = None

    #

    __repr__ = AttrOps['HttpClientRequest'](lambda o: (
        o.url,
        o.method,
        (o.headers, dict(repr_fn=AttrOps.truthy_repr)),
        (o.data, dict(repr_fn=lambda v: '...' if v is not None else None)),
        (o.timeout_s, dict(repr_fn=AttrOps.opt_repr)),
    )).repr

    #

    @property
    def method_or_default(self) -> str:
        if self.method is not None:
            return self.method
        if self.data is not None:
            return 'POST'
        return 'GET'

    @cached_property
    def headers_(self) -> ta.Optional[HttpHeaders]:
        return HttpHeaders(self.headers) if self.headers is not None else None


#


@dc.dataclass(frozen=True)  # kw_only=True
class BaseHttpClientResponse(Abstract):
    request: HttpClientRequest

    status: int

    headers: ta.Optional[HttpHeaders] = None

    underlying: ta.Any = None

    #

    __repr__ = AttrOps['BaseHttpClientResponse'](lambda o: (
        o.status,
        (o.headers, dict(repr_fn=AttrOps.truthy_repr)),
        o.request,
    )).repr

    #

    @property
    def is_success(self) -> bool:
        return is_success_http_status(self.status)


@ta.final
@dc.dataclass(frozen=True)  # kw_only=True
class HttpClientResponse(BaseHttpClientResponse):
    data: ta.Optional[bytes] = None

    #

    __repr__ = AttrOps['HttpClientResponse'](lambda o: (
        o.status,
        (o.headers, dict(repr_fn=AttrOps.truthy_repr)),
        (o.data, dict(repr_fn=lambda v: '...' if v is not None else None)),
        o.request,
    )).repr


##


@ta.final
class HttpClientContext:
    def __init__(self) -> None:
        self._dct: dict = {}


##


class HttpClientError(Exception):
    @property
    def cause(self) -> ta.Optional[BaseException]:
        return self.__cause__


@dc.dataclass()
class StatusHttpClientError(HttpClientError):
    response: HttpClientResponse


##


class BaseHttpClient(Abstract):
    pass
